from __future__ import annotations

from typing import final

from msgspec import field

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .message import Message
from .user import User


@final
class CallbackQuery(TelegramType, kw_only=True):
    """This object represents an incoming callback query from a callback button
    in an inline keyboard. If the button that originated the query was attached
    to a message sent by the bot, the field ``message`` will be present. If the
    button was attached to a message sent via the bot (in inline mode), the
    field ``inline_message_id`` will be present. Exactly one of the fields data
    or ``game_short_name`` will be present.

    See: https://core.telegram.org/bots/api#callbackquery
    """

    id: str
    """Unique identifier for this query"""

    from_: User = field(name="from")
    """Sender"""

    message: Message | None = None
    """*Optional.* Message with the callback button that originated the query.
    Note that message content and message date will not be available if the
    message is too old.
    """

    inline_message_id: NoneStr = None
    """*Optional.* Identifier of the message sent via the bot in inline mode,
    that originated the query.
    """

    chat_instance: str
    """Global identifier, uniquely corresponding to the chat to which the
    message with the callback button was sent. Useful for high scores in games.
    """

    data: NoneStr = None
    """*Optional.* Data associated with the callback button. Be aware that the
    message originated the query can contain no callback buttons with this data.
    """

    game_short_name: NoneStr = None
    """*Optional.* Short name of a `Game`_ to be returned, serves as the unique
    identifier for the game.

    .. _game: https://core.telegram.org/bots/api#games
    """

    async def answer(
        self,
        text: NoneStr = None,
        show_alert: NoneBool = None,
        url: NoneStr = None,
        cache_time: NoneInt = None,
    ) -> bool:
        """Use this method to send answer to callback.

        See: :meth:`Bot.answer_callback_query <yatbaf.bot.Bot.answer_callback_query>`
        """  # noqa: E501

        return await self.bot.answer_callback_query(
            callback_query_id=self.id,
            text=text,
            show_alert=show_alert,
            url=url,
            cache_time=cache_time,
        )



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class ReopenForumTopic(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.reopen_forum_topic`"""

    chat_id: str | int
    message_thread_id: int



import pytest

from yatbaf.enums import AdminFlag
from yatbaf.enums import MarkdownEntity
from yatbaf.helpers import create_bot_deeplink
from yatbaf.helpers import create_channel_deeplink
from yatbaf.helpers import create_game_deeplink
from yatbaf.helpers import create_group_deeplink
from yatbaf.helpers import create_user_link
from yatbaf.helpers import create_webapp_deeplink
from yatbaf.helpers import md
from yatbaf.helpers import parse_command_args


@pytest.mark.parametrize(
    "text,result",
    (
        ("/foo", []),
        ("/foo bar", ["bar"]),
        ("/foo@bot bar baz", ["bar", "baz"]),
        ("/foo Bar baZ 123", ["Bar", "baZ", "123"]),
    ),
)
def test_parse_args(text, result):
    assert parse_command_args(text) == result


@pytest.mark.parametrize(
    "text,expect",
    (
        ("foo", "foo"),
        ("foo.", "foo\\."),
        ("*foo*.", "\\*foo\\*\\."),
        ("_*foo__.", "\\_\\*foo\\_\\_\\."),
        ("__foo__", "\\_\\_foo\\_\\_"),
        ("!foo__.", "\\!foo\\_\\_\\."),
        ("foo != bar", "foo \\!\\= bar"),
        ("foo|bar", "foo\\|bar"),
        ("foo {} bar", "foo \\{\\} bar"),
        ("foo > bar", "foo \\> bar"),
        ("#!/usr/foo/bar -baz", "\\#\\!/usr/foo/bar \\-baz"),
        ("#![!_foo(~bar_)`].", "\\#\\!\\[\\!\\_foo\\(\\~bar\\_\\)\\`\\]\\."),
    )
)
def test_markdown_escape(text, expect):
    assert (result := md.escape(text)) == expect, result


@pytest.mark.parametrize(
    "text,expect", (
        ("foo", "foo"),
        ("foo {}", "foo {}"),
        ("foo {0}", "foo {0}"),
    )
)
def test_markdown_escape_fstring(text, expect):
    assert (result := md.escape(text, MarkdownEntity.FSTRING)) == expect, result
    assert result.format(s := "bar") == expect.format(s)


@pytest.mark.parametrize(
    "text,expect",
    (
        ("foo", "foo"),
        ("`foo`", "\\`foo\\`"),
        ("``\\foo\\``", "\\`\\`\\\\foo\\\\\\`\\`"),
    )
)
def test_markdown_escape_code_pre(text, expect):
    assert md.escape(text, MarkdownEntity.PRE) == expect
    assert md.escape(text, MarkdownEntity.CODE) == expect


@pytest.mark.parametrize(
    "text,expect",
    (
        ("foo", "foo"),
        ("http://foo.bar?q=(\\baz)", "http://foo.bar?q=(\\\\baz\\)"),
        ("http://foo.bar?q=\\))\\!baz", "http://foo.bar?q=\\\\\\)\\)\\\\!baz"),
    )
)
def test_markdown_escape_link_emoji(text, expect):
    assert md.escape(text, MarkdownEntity.LINK) == expect
    assert md.escape(text, MarkdownEntity.EMOJI) == expect


def test_markdown_bold():
    assert md.bold("bold") == "*bold*"


def test_markdown_italic():
    assert md.italic("italic") == "_italic_"


def test_markdown_underline():
    assert md.underline("underline") == "__underline__"


def test_markdown_strikethrough():
    assert md.strikethrough("strikethrough") == "~strikethrough~"


def test_markdown_spoiler():
    assert md.spoiler("spoiler") == "||spoiler||"


def test_markdown_url():
    assert (
        md.url("url title", "https://foo.bar") == "[url title](https://foo.bar)"
    )


def test_markdown_mention():
    assert md.mention("user", 1234) == "[user](tg://user?id=1234)"


def test_markdown_emoji():
    assert md.emoji(1234, "emoji") == "![emoji](tg://emoji?id=1234)"


def test_markdown_inline():
    assert md.inline("foo") == "`foo`"


def test_markdown_code():
    assert md.code("print('foo')") == "```\nprint('foo')\n```"
    assert md.code("print('foo')", "python") == "```python\nprint('foo')\n```"


@pytest.mark.parametrize("username", ("bot123", "user_bot"))
def test_create_user_link(username):
    assert create_user_link(username) == f"https://t.me/{username}"


@pytest.mark.parametrize("username", ("bot-123", "bt", "@bot"))
def test_create_user_link_error(username):
    with pytest.raises(ValueError):
        create_user_link(username)


@pytest.mark.parametrize("parameter", ("123-payload", None))
def test_create_bot_deeplink(parameter):
    bot_username = "username_bot"
    result = create_bot_deeplink(bot_username, parameter)
    if parameter is not None:
        assert result == f"https://t.me/{bot_username}?start={parameter}"
    else:
        assert result == f"https://t.me/{bot_username}"


def test_create_bot_deeplink_parameter_gt64():
    with pytest.raises(ValueError):
        create_bot_deeplink("botusername", "parameter" * 10)


@pytest.mark.parametrize("parameter", (
    "param@",
    "param&",
    ";;;",
    "foo bar",
))
def test_create_bot_deeplink_parameter_invalid_chars(parameter):
    with pytest.raises(ValueError):
        create_bot_deeplink("botusername", parameter)


def test_create_group_deeplink_empty():
    bot = "botusername"
    assert create_group_deeplink(bot) == f"https://t.me/{bot}?startgroup"


def test_create_group_deeplink_parameter():
    bot = "botusername"
    parameter = "parameter_123"
    assert (
        create_group_deeplink(bot, parameter) ==
        f"https://t.me/{bot}?startgroup={parameter}"
    )


def test_create_group_deeplink_parameter_empty():
    bot = "botusername"
    assert create_group_deeplink(bot, "") == f"https://t.me/{bot}?startgroup"


def test_create_group_deeplink_admin():
    bot = "botusername"
    flags = [AdminFlag.CHANGE_INFO, AdminFlag.EDIT_MESSAGES]
    assert (
        create_group_deeplink(bot, admin=flags) ==
        f"https://t.me/{bot}?startgroup&admin=change_info+edit_messages"
    )


def test_create_group_deeplink_admin_empty():
    bot = "botusername"
    assert (
        create_group_deeplink(bot, admin=[]) == f"https://t.me/{bot}?startgroup"
    )


def test_create_group_deeplink_parameter_admin():
    bot = "botusername"
    parameter = "param_foo-bar"
    flags = [
        AdminFlag.EDIT_MESSAGES,
        AdminFlag.INVITE_USERS,
        AdminFlag.DELETE_MESSAGES,
    ]
    assert (
        create_group_deeplink(bot, parameter, flags) == (
            f"https://t.me/{bot}?startgroup={parameter}"
            "&admin=edit_messages+invite_users+delete_messages"
        )
    )


def test_create_group_deeplink_parameter_error():
    with pytest.raises(ValueError):
        create_group_deeplink("botusername", "parameter#")


def test_create_channel_deeplink():
    bot = "botusername"
    flags = [
        AdminFlag.EDIT_MESSAGES,
        AdminFlag.DELETE_MESSAGES,
    ]
    assert (
        create_channel_deeplink(bot, flags) == (
            f"https://t.me/{bot}?startchannel"
            "&admin=edit_messages+delete_messages"
        )
    )


def test_create_channel_deeplink_admin_empty():
    with pytest.raises(ValueError):
        create_channel_deeplink("botusername", [])


def test_create_game_deeplink():
    bot = "botusername"
    game = "gameshortname"
    assert create_game_deeplink(bot, game) == f"https://t.me/{bot}?game={game}"


def test_create_game_deeplink_error():
    with pytest.raises(ValueError):
        create_game_deeplink("invalidbot#", "gameshortname")


def test_create_webapp_deeplink():
    bot = "botusername"
    webapp = "webappname"
    assert (
        create_webapp_deeplink(bot, webapp) ==
        f"https://t.me/{bot}/{webapp}?startapp"
    )


def test_create_webapp_deeplink_parameter():
    bot = "botusername"
    webapp = "webappname"
    parameter = "parameter"
    assert (
        create_webapp_deeplink(bot, webapp, parameter) ==
        f"https://t.me/{bot}/{webapp}?startapp={parameter}"
    )


def test_create_webapp_deeplink_parameter_empty():
    bot = "botusername"
    webapp = "webappname"
    parameter = ""
    assert (
        create_webapp_deeplink(bot, webapp, parameter) ==
        f"https://t.me/{bot}/{webapp}?startapp"
    )


def test_create_webapp_deeplink_parameter_error():
    with pytest.raises(ValueError):
        create_webapp_deeplink("botusername", "webapp", "invalidparam#")



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class HideGeneralForumTopic(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.hide_general_forum_topic`"""

    chat_id: str | int



import unittest.mock as mock

import pytest

from yatbaf.enums import UpdateType
from yatbaf.filters import Chat
from yatbaf.filters import Command
from yatbaf.filters import User
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def test_new_hander(handler):
    h = Handler(handler, update_type=UpdateType.MESSAGE)
    assert not h._filters
    assert not h._middleware
    assert h._update_type is UpdateType.MESSAGE
    assert str(h) == "<Handler[type=message]>"
    assert h._match_fn is all


def test_filter(handler):
    cmd_filter = Command("cmd")
    chat_filter = Chat()
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[cmd_filter, chat_filter],
    )
    assert len(h._filters) == 2
    assert h._filters[0] is chat_filter
    assert h._filters[1] is cmd_filter


def test_filter_sort_false(handler):
    cmd_filter = Command("cmd")
    chat_filter = Chat()
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[cmd_filter, chat_filter],
        sort_filters=False,
    )
    assert len(h._filters) == 2
    assert h._filters[0] is cmd_filter
    assert h._filters[1] is chat_filter


def test_middleware(handler):
    middleware = mock.Mock()
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        middleware=[middleware],
    )
    assert middleware in h._middleware


def test_wrap_middlewares(handler):
    middleware = mock.Mock(return_value=(func := mock.AsyncMock()))
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        middleware=[middleware],
    )
    assert h._build_middleware_stack() is func


def test_wrap_middlewares_empty(handler):
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
    )
    assert h._build_middleware_stack() is handler


def test_wrap_middleware_order(handler):
    middleware1 = mock.Mock(return_value=(func1 := mock.AsyncMock()))
    middleware2 = mock.Mock()
    middleware3 = mock.Mock()
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        middleware=[
            middleware1,
            middleware2,
            middleware3,
        ],
    )
    assert h._build_middleware_stack() is func1


def test_wrap_middleware_parent(handler):
    middleware1 = mock.Mock(return_value=(func1 := mock.AsyncMock()))
    middleware2 = mock.Mock()
    middleware3 = mock.Mock()
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        middleware=[
            middleware2,
            middleware3,
        ],
    )
    _ = OnMessage(
        middleware=[middleware1],
        handlers=[h],
    )
    assert h._build_middleware_stack() is func1


def test_match_empty_true(handler, message):
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
    )
    assert h._match(message)


def test_is_fallback(handler):
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    assert not h._is_fallback()


def test_empty_is_fallback(handler):
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
    )
    assert h._is_fallback()


def test_match_true(handler, message):
    message.text = "/foo"
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    assert h._match(message)


def test_match_true1(handler, message):
    message.text = "/foo"
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[
            Command("foo"),
            User(message.from_.id),
        ]
    )
    assert h._match(message)


def test_match_false(handler, message):
    message.text = "/bar"
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    assert not h._match(message)


def test_match_false1(handler, message):
    message.text = "/foo"
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
        filters=[
            Command("bar"),
            User(1),
        ],
    )
    assert not h._match(message)


@pytest.mark.asyncio
async def test_orig_func(handler, message):
    h = Handler(
        handler,
        update_type=UpdateType.MESSAGE,
    )
    await h.orig(message)
    handler.assert_awaited_once_with(message)



import re
import unittest.mock as mock

import pytest

from yatbaf.enums import ChatType
from yatbaf.enums import ContentType
from yatbaf.filters import And
from yatbaf.filters import Channel
from yatbaf.filters import Command
from yatbaf.filters import Content
from yatbaf.filters import Group
from yatbaf.filters import Not
from yatbaf.filters import Or
from yatbaf.filters import Text
from yatbaf.filters import User

USER_ID = 34567
CHAT_ID = 9876


@pytest.fixture
def update_message():
    update = mock.MagicMock()
    update.from_.id = USER_ID
    update.from_.username = "test_user"
    update.chat.id = CHAT_ID
    return update


class FalseFilter:
    priority = 100

    def check(self, _):  # noqa: U101
        return False


class TrueFilter:
    priority = 100

    def check(self, _):  # noqa: U101
        return True


def test_filter_not_true(update_message):
    assert Not(FalseFilter()).check(update_message)


def test_filter_not_false(update_message):
    assert not Not(TrueFilter()).check(update_message)


def test_filter_or_true(update_message):
    assert Or(FalseFilter(), TrueFilter()).check(update_message)


def test_filter_or_false(update_message):
    assert not Or(FalseFilter(), FalseFilter()).check(update_message)


def test_filter_and_true(update_message):
    assert And(TrueFilter(), TrueFilter()).check(update_message)


def test_filter_and_false(update_message):
    assert not And(TrueFilter(), FalseFilter()).check(update_message)


@pytest.mark.parametrize("user", ("@test_user", "test_user", USER_ID))
def test_filter_user(update_message, user):
    assert User(user).check(update_message)


@pytest.mark.parametrize("user", ("@testuser", "test_user1", 23345678))
def test_filter_not_user(update_message, user):
    assert not User(user).check(update_message)


@pytest.mark.parametrize("user", ("@test_user", "test_user", USER_ID))
def test_filter_invert_user(update_message, user):
    assert not Not(User(user)).check(update_message)


@pytest.mark.parametrize("t", (ChatType.GROUP, ChatType.SUPERGROUP))
def test_filter_group(update_message, t):
    update_message.chat.type = t
    assert Group().check(update_message)


@pytest.mark.parametrize("t", (ChatType.GROUP, ChatType.SUPERGROUP))
def test_filter_group_ids(update_message, t):
    update_message.chat.type = t
    assert Group(CHAT_ID).check(update_message)
    assert not Group(123).check(update_message)


def test_filter_channel(update_message):
    update_message.chat.type = ChatType.CHANNEL
    assert Channel().check(update_message)


def test_filter_channel_ids(update_message):
    update_message.chat.type = ChatType.CHANNEL
    assert Channel(CHAT_ID).check(update_message)
    assert not Channel(123).check(update_message)


def test_filter_content_empty():
    with pytest.raises(ValueError):
        Content()


def test_filter_content_wrong_type():
    with pytest.raises(ValueError):
        Content("typo")


def test_filter_content_true(update_message):
    update_message.text = "123"
    assert Content("text").check(update_message)
    assert Content(ContentType.TEXT).check(update_message)


def test_filter_content_false(update_message):
    update_message.text = "123"
    update_message.document = None
    assert not Content("document").check(update_message)
    assert not Content(ContentType.DOCUMENT).check(update_message)


def test_filter_command_empty():
    with pytest.raises(ValueError):
        Command()


def test_filter_command_true(update_message):
    update_message.text = "/start"
    assert Command("start").check(update_message)
    assert Command("/start").check(update_message)
    assert Command("START").check(update_message)
    assert Command("Start").check(update_message)
    assert Command("hello", "start").check(update_message)


def test_filter_command_false(update_message):
    update_message.text = "/pong"
    assert not Command("ping").check(update_message)


def test_filter_command_text_is_none(update_message):
    update_message.text = None
    assert not Command("ping").check(update_message)


def test_filter_text_no_params():
    with pytest.raises(ValueError):
        Text()


def test_filter_text_is_text(update_message):
    update_message.text = None
    assert not Text(startswith="foo").check(update_message)


@pytest.mark.parametrize("f,ic", (("FOO", False), ("foo", True)))
def test_filter_text_start(update_message, f, ic):
    update_message.text = "FOO bar"
    assert Text(startswith=f, ignore_case=ic).check(update_message)


@pytest.mark.parametrize("f,ic", (("BAR", False), ("bar", True)))
def test_filter_text_end(update_message, f, ic):
    update_message.text = "foo BAR"
    assert Text(endswith=f, ignore_case=ic).check(update_message)


def test_filter_text_start_end(update_message):
    update_message.text = "foo baz bar"
    assert Text(startswith="foo", endswith="bar").check(update_message)


def test_filter_text_start_end_any(update_message):
    update_message.text = "foo bar baz"
    assert (
        Text(startswith="foo", endswith="bar", any_=True).check(update_message)
    )


@pytest.mark.parametrize("m,ic", (("Foo Bar", False), ("foo bar", True)))
def test_filter_text_match(update_message, m, ic):
    update_message.text = "Foo Bar"
    assert Text(match=m, ignore_case=ic).check(update_message)


def test_filter_text_match_false(update_message):
    update_message.text = "foo bar"
    assert not Text(match="foo b").check(update_message)


@pytest.mark.parametrize("c", ("bar", ["oof", "foo"]))
def test_filter_text_contains(update_message, c):
    update_message.text = "foo bar baz"
    assert Text(contains=c).check(update_message)


def test_filter_text_contains_false(update_message):
    update_message.text = "foo bar baz"
    assert not Text(contains="foobar").check(update_message)


@pytest.mark.parametrize("t", ("foobarbaz", "foo bar baz"))
def test_filter_text_start_end_contains(update_message, t):
    update_message.text = t
    assert Text(
        startswith="fo",
        endswith="az",
        contains="bar",
    ).check(update_message)


@pytest.mark.parametrize("exp", (".+baz.*$", re.compile("^f.+az$")))
def test_filter_text_regexp(update_message, exp):
    update_message.text = "foo bar baz"
    assert Text(regexp=exp).check(update_message)


def test_priority():
    user = User(123)
    channel = Channel()
    command = Command("start")
    content = Content("text")

    assert Not(channel).priority == channel.priority
    assert And(command, user).priority == user.priority
    assert Or(channel, content).priority == channel.priority



from yatbaf.bot import Bot
from yatbaf.methods import __all__ as methods

# Bot API 6.9
API_METHODS = (
    "get_updates",
    "set_webhook",
    "delete_webhook",
    "get_webhook_info",
    "get_me",
    "log_out",
    "close",
    "send_message",
    "forward_message",
    "copy_message",
    "send_photo",
    "send_audio",
    "send_document",
    "send_video",
    "send_animation",
    "send_voice",
    "send_video_note",
    "send_media_group",
    "send_location",
    "send_venue",
    "send_contact",
    "send_poll",
    "send_dice",
    "send_chat_action",
    "get_user_profile_photos",
    "get_file",
    "ban_chat_member",
    "unban_chat_member",
    "restrict_chat_member",
    "promote_chat_member",
    "set_chat_administrator_custom_title",
    "ban_chat_sender_chat",
    "unban_chat_sender_chat",
    "set_chat_permissions",
    "export_chat_invite_link",
    "create_chat_invite_link",
    "edit_chat_invite_link",
    "revoke_chat_invite_link",
    "approve_chat_join_request",
    "decline_chat_join_request",
    "set_chat_photo",
    "delete_chat_photo",
    "set_chat_title",
    "set_chat_description",
    "pin_chat_message",
    "unpin_chat_message",
    "unpin_all_chat_messages",
    "leave_chat",
    "get_chat",
    "get_chat_administrators",
    "get_chat_member_count",
    "get_chat_member",
    "set_chat_sticker_set",
    "delete_chat_sticker_set",
    "get_forum_topic_icon_stickers",
    "create_forum_topic",
    "edit_forum_topic",
    "close_forum_topic",
    "reopen_forum_topic",
    "delete_forum_topic",
    "unpin_all_forum_topic_messages",
    "edit_general_forum_topic",
    "close_general_forum_topic",
    "reopen_general_forum_topic",
    "hide_general_forum_topic",
    "unhide_general_forum_topic",
    "answer_callback_query",
    "set_my_name",
    "get_my_name",
    "set_my_commands",
    "delete_my_commands",
    "get_my_commands",
    "set_my_description",
    "get_my_description",
    "set_my_short_description",
    "get_my_short_description",
    "set_chat_menu_button",
    "get_chat_menu_button",
    "set_my_default_administrator_rights",
    "get_my_default_administrator_rights",
    "edit_message_text",
    "edit_message_caption",
    "edit_message_media",
    "edit_message_live_location",
    "stop_message_live_location",
    "edit_message_reply_markup",
    "stop_poll",
    "delete_message",
    "send_sticker",
    "get_sticker_set",
    "get_custom_emoji_stickers",
    "upload_sticker_file",
    "create_new_sticker_set",
    "add_sticker_to_set",
    "set_sticker_position_in_set",
    "delete_sticker_from_set",
    "set_sticker_emoji_list",
    "set_sticker_keywords",
    "set_sticker_mask_position",
    "set_sticker_set_title",
    "set_sticker_set_thumbnail",
    "set_custom_emoji_sticker_set_thumbnail",
    "delete_sticker_set",
    "answer_inline_query",
    "answer_web_app_query",
    "send_invoice",
    "create_invoice_link",
    "answer_shipping_query",
    "answer_pre_checkout_query",
    "set_passport_data_errors",
    "send_game",
    "set_game_score",
    "get_game_high_scores",
    "unpin_all_general_forum_topic_messages",
)


def test_bot_api_methods():
    for method in API_METHODS:
        assert hasattr(Bot, method)


def test_api_methods():
    assert len(methods) == len(API_METHODS)
    m1 = map(lambda x: x.replace("_", ""), API_METHODS)
    m2 = map(lambda x: x.lower(), methods)
    assert sorted(m1) == sorted(m2)



import unittest.mock as mock

import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.enums import UpdateType
from yatbaf.handler import Handler
from yatbaf.middleware import Middleware
from yatbaf.models import UpdateInfo
from yatbaf.router import OnEditedChannelPost
from yatbaf.router import OnEditedMessage
from yatbaf.router import OnMessage
from yatbaf.router import OnPoll


def test_routers(token):
    dispatcher = Dispatcher(token)
    for type_ in UpdateType:
        dispatcher._routers[type_]


def test_dispatcher(token):
    dispatcher = Dispatcher(token)
    assert dispatcher._routers
    assert not dispatcher._middleware
    assert not dispatcher._guards


def test_add_router(token, router):
    dispatcher = Dispatcher(token, routers=[router])
    assert len(routers := dispatcher._routers[router._update_type]) == 1
    assert router in routers


def test_add_router_duplicate(token):
    router = OnMessage()
    dispatcher = Dispatcher(token, routers=[router, router])
    assert len(routers := dispatcher._routers[router._update_type]) == 1
    assert router in routers


def test_add_routers(token):
    dispatcher = Dispatcher(
        token,
        routers=[
            message := OnMessage(),
            poll := OnPoll(),
            edited_message := OnEditedMessage(),
            edited_post := OnEditedChannelPost(),
        ],
    )
    assert message in dispatcher._routers["message"]
    assert poll in dispatcher._routers["poll"]
    assert edited_message in dispatcher._routers["edited_message"]
    assert edited_post in dispatcher._routers["edited_channel_post"]


def test_dispatcher_update_type(token):
    with pytest.raises(RuntimeError):
        Dispatcher(token)._update_type


def test_middleware(token):
    fn = mock.Mock(return_value=(mdwl := mock.Mock))
    dispatcher = Dispatcher(token, middleware=[Middleware(fn)])
    assert not dispatcher._middleware
    assert dispatcher._middleware_stack is mdwl


def test_middleware_order(token):
    fn1 = mock.Mock(return_value=(mdwl1 := mock.Mock))
    fn2 = mock.Mock()
    dispatcher = Dispatcher(
        token, middleware=[
            Middleware(fn1),
            Middleware(fn2),
        ]
    )
    assert not dispatcher._middleware
    assert dispatcher._middleware_stack is mdwl1


@pytest.mark.asyncio
async def test_resolve_no_routers(token, update_info):
    dispatcher = Dispatcher(token)
    assert not await dispatcher._resolve(update_info)


@pytest.mark.asyncio
async def test_resolve(token, handler, update_info):
    dispatcher = Dispatcher(
        token,
        routers=[OnMessage(handlers=[handler])],
    )
    await dispatcher._resolve(update_info)
    handler.assert_awaited_once_with(update_info.content)


@pytest.mark.asyncio
async def test_resolve_none(token, handler, update_info):
    dispatcher = Dispatcher(token, routers=[OnPoll(handlers=[handler])])
    await dispatcher._resolve(update_info)
    handler.assert_not_awaited()


@pytest.mark.asyncio
async def test_resolve_guard_false(token, handler, update_info, asyncdef):
    dispatcher = Dispatcher(
        token,
        guards=[asyncdef(False)],
        routers=[OnMessage(handlers=[handler])]
    )
    await dispatcher._resolve(update_info)
    handler.assert_not_awaited()


@pytest.mark.asyncio
async def test_resolve_guard_true(token, handler, update_info, asyncdef):
    dispatcher = Dispatcher(
        token,
        guards=[asyncdef(True)],
        routers=[OnMessage(handlers=[handler])],
    )
    await dispatcher._resolve(update_info)
    handler.assert_awaited_once_with(update_info.content)


@pytest.mark.asyncio
async def test_process_update(monkeypatch, token, update):
    monkeypatch.setattr(Dispatcher, "_bind_self", lambda _, v: v)
    monkeypatch.setattr(
        Dispatcher, "_resolve", resolve := mock.AsyncMock(return_value=None)
    )
    dispatcher = Dispatcher(token)
    await dispatcher.process_update(update)
    resolve.assert_awaited_once_with(
        UpdateInfo(
            id=update.update_id,
            content=update.message,
            name="message",
        )
    )


def test_add_handler_duplicate(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            Handler(handler, UpdateType.MESSAGE),
            Handler(handler, UpdateType.MESSAGE),
        ],
    )
    assert dispatcher._routers["message"]
    assert len(dispatcher._routers["message"]) == 1


def test_add_handler_order(token):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h1 := Handler(object(), UpdateType.MESSAGE),
            h3 := Handler(object(), UpdateType.MESSAGE),
            h2 := Handler(object(), UpdateType.MESSAGE),
        ]
    )
    assert len(dispatcher._routers["message"]) == 1
    assert len(dispatcher._routers["message"][-1]._handlers) == 3
    assert dispatcher._routers["message"][-1]._handlers[0] is h1
    assert dispatcher._routers["message"][-1]._handlers[1] is h3
    assert dispatcher._routers["message"][-1]._handlers[2] is h2


def test_add_handler_message(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.MESSAGE),
        ],
    )
    assert dispatcher._routers["message"][-1]._handlers[0] is h


def test_add_handler_edited_message(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.EDITED_MESSAGE),
        ],
    )
    assert dispatcher._routers["edited_message"][-1]._handlers[0] is h


def test_add_handler_channel_post(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.CHANNEL_POST),
        ],
    )
    assert dispatcher._routers["channel_post"][-1]._handlers[0] is h


def test_add_handler_edited_channel_post(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.EDITED_CHANNEL_POST),
        ],
    )
    assert dispatcher._routers["edited_channel_post"][-1]._handlers[0] is h


def test_add_handler_inline_query(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.INLINE_QUERY),
        ],
    )
    assert dispatcher._routers["inline_query"][-1]._handlers[0] is h


def test_add_handler_chosen_inline_result(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.CHOSEN_INLINE_RESULT),
        ],
    )
    assert dispatcher._routers["chosen_inline_result"][-1]._handlers[0] is h


def test_add_handler_callback_query(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.CALLBACK_QUERY),
        ],
    )
    assert dispatcher._routers["callback_query"][-1]._handlers[0] is h


def test_add_handler_shipping_query(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.SHIPPING_QUERY),
        ],
    )
    assert dispatcher._routers["shipping_query"][-1]._handlers[0] is h


def test_add_handler_pre_checkout_query(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.PRE_CHECKOUT_QUERY),
        ],
    )
    assert dispatcher._routers["pre_checkout_query"][-1]._handlers[0] is h


def test_add_handler_poll(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.POLL),
        ],
    )
    assert dispatcher._routers["poll"][-1]._handlers[0] is h


def test_add_handler_poll_answer(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.POLL_ANSWER),
        ],
    )
    assert dispatcher._routers["poll_answer"][-1]._handlers[0] is h


def test_add_handler_my_chat_member(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.MY_CHAT_MEMBER),
        ],
    )
    assert dispatcher._routers["my_chat_member"][-1]._handlers[0] is h


def test_add_handler_chat_member(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.CHAT_MEMBER),
        ],
    )
    assert dispatcher._routers["chat_member"][-1]._handlers[0] is h


def test_add_handler_chat_join_request(token, handler):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler, UpdateType.CHAT_JOIN_REQUEST),
        ],
    )
    assert dispatcher._routers["chat_join_request"][-1]._handlers[0] is h



import pytest


def test_attr_is_private(message):
    assert "__usrctx__" not in message.__struct_fields__
    assert "__usrctx__" not in message.__slots__


def test_ctx_empty(message):
    assert not message.ctx


def test_ctx(message):
    message.ctx["foo"] = "bar"
    assert "foo" in message.__usrctx__


def test_bind_bot_obj(message):
    bot = object()
    message._bind_bot_obj(bot)
    assert message.bot is bot
    assert message.from_.bot is bot
    assert message.chat.bot is bot


def test_bot_not_bound(message):
    with pytest.raises(RuntimeError):
        message.bot



from __future__ import annotations

from typing import final

from msgspec import field

from yatbaf.enums import ChatType

from .abc import TelegramType
from .location import Location
from .user import User


@final
class InlineQuery(TelegramType):
    """This object represents an incoming inline query. When the user sends an
    empty query, your bot could return some default or trending results.

    See: https://core.telegram.org/bots/api#inlinequery
    """

    id: str
    """Unique identifier for this query."""

    from_: User = field(name="from")
    """Sender"""

    query: str
    """Text of the query (up to 256 characters)"""

    offset: str
    """Offset of the results to be returned, can be controlled by the bot"""

    chat_type: ChatType | None = None
    """*Optional.* Type of the chat from which the inline query was sent.

    .. note::

        The chat type should be always known for requests sent from official
        clients and most third-party clients, unless the request was sent from
        a secret chat.
    """

    location: Location | None = None
    """*Optional.* Sender location, only for bots that request user location."""



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class UserShared(TelegramType):
    """This object contains information about the user whose identifier was
    shared with the bot using a :class:`KeyboardButtonRequestUser <yatbaf.types.keyboard_button.KeyboardButtonRequestUser>` button.

    See: https://core.telegram.org/bots/api#usershared
    """  # noqa: E501

    request_id: int
    """Identifier of the request."""

    user_id: int
    """Identifier of the shared user."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.enums import PollType
    from yatbaf.types import MessageEntity
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr
    from yatbaf.typing import ReplyMarkup

# from datetime import timedelta


@final
class SendPoll(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_poll`"""

    chat_id: str | int
    question: str
    options: list[str]
    message_thread_id: NoneInt = None
    is_anonymous: NoneBool = None
    type: PollType | None = None
    allows_multiple_answers: NoneBool = None
    correct_option_id: NoneInt = None
    explanation: NoneStr = None
    explanation_parse_mode: ParseMode | None = None
    explanation_entities: list[MessageEntity] | None = None
    open_period: NoneInt = None
    close_date: int | None = None  # timedelta
    is_closed: NoneBool = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_to_message_id: NoneInt = None
    allow_sending_without_reply: NoneBool = None
    reply_markup: ReplyMarkup | None = None



import pytest
from msgspec import json as jsonlib

from yatbaf.enums import ParseMode
from yatbaf.methods import Close
from yatbaf.methods import EditMessageMedia
from yatbaf.methods import GetMe
from yatbaf.methods import GetUpdates
from yatbaf.methods import SendAnimation
from yatbaf.methods import SendAudio
from yatbaf.methods import SendDocument
from yatbaf.methods import SendMediaGroup
from yatbaf.methods import SendMessage
from yatbaf.methods import SendPhoto
from yatbaf.methods import SendSticker
from yatbaf.methods import SendVideo
from yatbaf.methods import SendVideoNote
from yatbaf.methods import SendVoice
from yatbaf.methods import SetWebhook
from yatbaf.methods import StopMessageLiveLocation
from yatbaf.methods import UploadStickerFile
from yatbaf.methods.abc import TelegramMethodWithFile
from yatbaf.methods.abc import TelegramMethodWithMedia
from yatbaf.types import InputMediaDocument
from yatbaf.types import InputMediaPhoto
from yatbaf.types import Message
from yatbaf.types import Update
from yatbaf.types import User


def test_base_subclass_attrs():
    assert not hasattr(TelegramMethodWithMedia, "__meth_name__")
    assert not hasattr(TelegramMethodWithMedia, "__meth_result_model__")

    assert not hasattr(TelegramMethodWithFile, "__meth_name__")
    assert not hasattr(TelegramMethodWithFile, "__meth_result_model__")


def test_attrs():
    assert hasattr(GetMe, "__meth_name__")
    assert hasattr(GetMe, "__meth_result_model__")


def test_file_fields():
    assert SendAnimation.__meth_file_fields__ == ("animation", "thumbnail")
    assert SendAudio.__meth_file_fields__ == ("audio",)
    assert SendDocument.__meth_file_fields__ == ("document",)
    assert SendPhoto.__meth_file_fields__ == ("photo",)
    assert SendSticker.__meth_file_fields__ == ("sticker",)
    assert SendVideo.__meth_file_fields__ == ("video", "thumbnail")
    assert SendVideoNote.__meth_file_fields__ == ("video_note", "thumbnail")
    assert SendVoice.__meth_file_fields__ == ("voice",)
    assert UploadStickerFile.__meth_file_fields__ == ("sticker",)
    assert SetWebhook.__meth_file_fields__ == ("certificate",)


def test_method_name():
    assert GetMe.__meth_name__ == "getme"
    assert str(GetMe()) == "getme"
    assert GetMe()._get_name() == "getme"


@pytest.mark.parametrize(
    "method,model",
    (
        (GetMe, User),
        (Close, bool),
        (GetUpdates, list[Update]),
        (StopMessageLiveLocation, Message | bool),
    ),
)
def test_method_result_model(method, model):
    assert method.__meth_result_model__ == model
    assert method._get_result_model() == model


def test_encode_empty():
    method = GetMe()
    data, files = method._encode_params()
    assert data is None
    assert files is None


def test_encode_params():
    method = SendMessage(
        chat_id=12345,
        text="*foo bar*",
        disable_notification=True,
        parse_mode=ParseMode.MARKDOWN,
    )
    data, files = method._encode_params()
    assert isinstance(data, bytes)
    assert files is None
    assert jsonlib.decode(data) == {
        "chat_id": 12345,
        "text": "*foo bar*",
        "disable_notification": True,
        "parse_mode": "MarkdownV2",
    }


def test_encode_file_id():
    method = SendPhoto(chat_id=123, photo="fileid")
    data, files = method._encode_params()
    assert files is None
    assert jsonlib.decode(data) == {"chat_id": 123, "photo": "fileid"}


def test_encode_file_content():
    photo = object()
    method = SendPhoto(chat_id=123, photo=photo)
    data, files = method._encode_params()

    assert files is not None
    assert isinstance(data, dict)
    assert data == {"chat_id": 123, "photo": (fn := method.photo)}
    assert files[fn.split("//")[1]] is photo


def test_encode_media_file_content():
    photo = object()
    method = EditMessageMedia(
        chat_id=12345,
        message_id=12345,
        media=InputMediaPhoto(media=photo),
    )
    data, files = method._encode_params()
    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "message_id": 12345,
        "media": {
            "media": (fn := method.media.media),
            "type": "photo",
        },
    }
    assert files[fn.split("//")[1]] is photo


def test_encode_media_filed_id():
    method = EditMessageMedia(
        chat_id=12345,
        message_id=12345,
        media=InputMediaPhoto(media="fileid"),
    )
    data, files = method._encode_params()

    assert files is None
    assert isinstance(data, bytes)

    assert jsonlib.decode(data) == {
        "chat_id": 12345,
        "message_id": 12345,
        "media": {
            "media": "fileid",
            "type": "photo",
        }
    }


def test_encode_media_group_file_content():
    photo = object()
    method = SendMediaGroup(
        media=[InputMediaPhoto(media=photo)],
        chat_id=12345,
    )
    data, files = method._encode_params()

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "media": [
            {
                "media": (fn := method.media[0].media),
                "type": "photo",
            },
        ],
    }
    assert files[fn.split("//")[1]] is photo


def test_encode_media_group_file_content_1():
    doc1 = object()
    doc2 = object()
    doc2_th = object()
    doc3 = object()

    method = SendMediaGroup(
        media=[
            InputMediaDocument(media=doc1),
            InputMediaDocument(media=doc2, thumbnail=doc2_th),
            InputMediaDocument(media=doc3),
        ],
        chat_id=12345,
    )
    data, files = method._encode_params()

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "media": [
            {
                "media": (fn1 := method.media[0].media),
                "type": "document",
            },
            {
                "media": (fn2 := method.media[1].media),
                "thumbnail": (fn2_th := method.media[1].thumbnail),
                "type": "document",
            },
            {
                "media": (fn3 := method.media[2].media),
                "type": "document",
            },
        ],
    }
    assert files[fn1.split("//")[1]] is doc1
    assert files[fn2.split("//")[1]] is doc2
    assert files[fn2_th.split("//")[1]] is doc2_th
    assert files[fn3.split("//")[1]] is doc3


def test_encode_media_group_file_id_content_mix():
    doc = object()
    doc_th = object()

    method = SendMediaGroup(
        media=[
            InputMediaDocument(media="fileid1"),
            InputMediaDocument(media="fileid2", thumbnail=doc_th),
            InputMediaDocument(media=doc),
        ],
        chat_id=12345,
    )
    data, files = method._encode_params()

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "media": [
            {
                "media": "fileid1",
                "type": "document",
            },
            {
                "media": "fileid2",
                "thumbnail": (fn2_th := method.media[1].thumbnail),
                "type": "document",
            },
            {
                "media": (fn3 := method.media[2].media),
                "type": "document",
            },
        ],
    }

    assert files[fn2_th.split("//")[1]] is doc_th
    assert files[fn3.split("//")[1]] is doc


def test_encode_media_group_file_id():
    method = SendMediaGroup(
        media=[
            InputMediaDocument(media="fileid1"),
            InputMediaDocument(media="fileid2", thumbnail="thid"),
            InputMediaDocument(media="fileid3"),
        ],
        chat_id=12345,
    )
    data, files = method._encode_params()

    assert files is None
    assert isinstance(data, bytes)
    assert jsonlib.decode(data) == {
        "chat_id": 12345,
        "media": [
            {
                "media": "fileid1",
                "type": "document",
            },
            {
                "media": "fileid2",
                "thumbnail": "thid",
                "type": "document",
            },
            {
                "media": "fileid3",
                "type": "document",
            },
        ],
    }



import unittest.mock as mock

import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.enums import UpdateType
from yatbaf.filters import Channel
from yatbaf.filters import Command
from yatbaf.filters import Content
from yatbaf.filters import User
from yatbaf.handler import Handler
from yatbaf.middleware import Middleware
from yatbaf.router import OnCallbackQuery
from yatbaf.router import OnChannelPost
from yatbaf.router import OnChatJoinRequest
from yatbaf.router import OnChatMemeber
from yatbaf.router import OnChosenInlineResult
from yatbaf.router import OnEditedChannelPost
from yatbaf.router import OnEditedMessage
from yatbaf.router import OnInlineQuery
from yatbaf.router import OnMessage
from yatbaf.router import OnMyChatMember
from yatbaf.router import OnPoll
from yatbaf.router import OnPollAnswer
from yatbaf.router import OnPreCheckoutQuery
from yatbaf.router import OnShippingQuery


@pytest.fixture
def dispatcher(token):
    return Dispatcher(token)


@pytest.fixture
def router():
    return OnMessage()


@pytest.fixture
def filter():
    return Command("foo")


@pytest.fixture
def handler_func():
    return mock.AsyncMock(return_value=None)


@pytest.fixture
def handler(handler_func):
    return Handler(handler_func, UpdateType.MESSAGE)


def test_update_type():
    assert OnCallbackQuery._update_type == UpdateType.CALLBACK_QUERY
    assert OnChannelPost._update_type == UpdateType.CHANNEL_POST
    assert OnChatJoinRequest._update_type == UpdateType.CHAT_JOIN_REQUEST
    assert OnChatMemeber._update_type == UpdateType.CHAT_MEMBER
    assert OnChosenInlineResult._update_type == UpdateType.CHOSEN_INLINE_RESULT
    assert OnEditedChannelPost._update_type == UpdateType.EDITED_CHANNEL_POST
    assert OnEditedMessage._update_type == UpdateType.EDITED_MESSAGE
    assert OnInlineQuery._update_type == UpdateType.INLINE_QUERY
    assert OnMessage._update_type == UpdateType.MESSAGE
    assert OnMyChatMember._update_type == UpdateType.MY_CHAT_MEMBER
    assert OnPoll._update_type == UpdateType.POLL
    assert OnPollAnswer._update_type == UpdateType.POLL_ANSWER
    assert OnPreCheckoutQuery._update_type == UpdateType.PRE_CHECKOUT_QUERY
    assert OnShippingQuery._update_type == UpdateType.SHIPPING_QUERY


def test_new_router():
    router = OnMessage()
    assert not router._handlers
    assert not router._middleware
    assert not router._guards


def test_add_handler(router, handler_func):
    router.add_handler(handler_func)
    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler_func


def test_add_handler_obj(router, handler_func):
    handler = Handler(handler_func, UpdateType.MESSAGE)
    router.add_handler(handler)
    assert router._handlers[0] is handler


def test_add_handler_type_error(router, handler_func):
    handler = Handler(handler_func, UpdateType.POLL)
    with pytest.raises(ValueError):
        router.add_handler(handler)


def test_add_handler_duplicate_func(router, handler_func, filter):
    handler1 = Handler(handler_func, UpdateType.MESSAGE)
    handler2 = Handler(handler_func, UpdateType.MESSAGE, filters=[filter])
    router.add_handler(handler1)
    router.add_handler(handler2)
    assert len(router._handlers) == 1
    assert handler1 in router._handlers
    assert handler2 not in router._handlers


def test_add_handler_duplicate_func1(router, handler_func, filter):
    handler1 = Handler(handler_func, UpdateType.MESSAGE)
    router.add_handler(handler1)
    router.add_handler(handler_func, filters=[filter])
    assert len(router._handlers) == 1
    assert handler1 in router._handlers


def test_add_handler_duplicate_func2(router, handler_func, filter):
    router.add_handler(handler_func)
    router.add_handler(handler_func, filters=[filter])
    assert len(router._handlers) == 1
    assert not router._handlers[0]._filters


def test_add_handler_decorator(router):

    @router
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler


def test_add_handler_decorator_duplicate(router):

    @router
    @router
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert len(router._handlers) == 1
    assert router._handlers[0]._fn is handler


def test_handler_decorator_filter(router):
    filter = Command("foo")

    @router(filters=[filter])
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler
    assert filter in router._handlers[0]._filters


def test_handler_decorator_middleware(router):
    middleware = mock.Mock()

    @router(middleware=[middleware])
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler
    assert middleware in router._handlers[0]._middleware


def test_handler_decorator_middleware_filter_mix(router, filter):
    middleware = mock.Mock()

    @router(
        middleware=[middleware],
        filters=[filter],
    )
    async def handler(_) -> None:  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler
    assert middleware in router._handlers[0]._middleware
    assert filter in router._handlers[0]._filters


def test_router_init_guard():
    func = object()
    router = OnMessage(guards=[func])
    assert func in router._guards


def test_router_add_guard():
    router = OnMessage()
    func = object()
    router.add_guard(func)
    assert len(router._guards) == 1
    assert func in router._guards


def test_router_add_guard_duplicate():
    router = OnMessage()
    func = object()

    router.add_guard(func)
    router.add_guard(func)
    assert len(router._guards) == 1
    assert func in router._guards


def test_router_guard_decorator(router):

    @router.guard
    async def func(_):  # noqa: U101
        pass

    assert len(router._guards) == 1
    assert func in router._guards


def test_router_init_middleware_handler():
    func = object()
    router = OnMessage(middleware=[func])
    assert Middleware(func, is_handler=True) in router._middleware


def test_router_init_middleware_router():
    func = object()
    router = OnMessage(middleware=[Middleware(func)])
    assert Middleware(func) in router._middleware


def test_router_init_middleware_handler_local():
    func = object()
    router = OnMessage(
        middleware=[
            Middleware(
                func,
                is_handler=True,
                is_local=True,
            ),
        ]
    )
    middleware = Middleware(
        func,
        is_handler=True,
        is_local=True,
    )
    assert middleware in router._middleware
    router._on_registration()
    assert middleware not in router._middleware


def test_router_add_middleware(router):
    func = object()
    router.add_middleware(func)
    assert Middleware(func, is_handler=True) in router._middleware


def test_router_middleware_handler_decorator(router):

    @router.middleware
    def func(_):  # noqa: U101
        pass

    assert len(router._middleware) == 1
    assert Middleware(func, is_handler=True) in router._middleware


def test_router_middleware_router_decorator(router):

    @router.middleware(is_handler=False)
    def func(_):  # noqa: U101
        pass

    assert len(router._middleware) == 1
    assert Middleware(func) in router._middleware


def test_router_middleware_decorator_duplicate(router):

    @router.middleware
    @router.middleware
    def func(_):  # noqa: U101
        pass

    assert len(router._middleware) == 1
    assert Middleware(func, is_handler=True) in router._middleware


def test_find_handler_none_no_handlers(router, message):
    assert router._find_handler(message) is None


def test_find_handler_no_filters(router, message, handler):
    router.add_handler(handler)
    assert router._find_handler(message) is handler


def test_find_handler_catch_all(handler, router, message, asyncdef, filter):
    router.add_handler(handler)
    router.add_handler(asyncdef(), filters=[filter])
    message.text = "/bar"
    assert router._find_handler(message) is handler


def test_find_handler_filter(router, message, asyncdef, filter, handler_func):
    router.add_handler(asyncdef())
    handler = Handler(handler_func, UpdateType.MESSAGE, filters=[filter])
    router.add_handler(handler)
    message.text = "/foo"
    assert router._find_handler(message) is handler


def test_add_router():
    router = OnMessage()
    nested = OnMessage()
    router.add_router(nested)
    assert len(router._routers) == 1
    assert nested in router._routers


def test_add_router_self(router):
    with pytest.raises(ValueError):
        router.add_router(router)


def test_add_router_wrong_type(router):
    with pytest.raises(ValueError):
        router.add_router(OnPoll())


def test_add_router_dispatcher(router, dispatcher):
    with pytest.raises(RuntimeError):
        router.add_router(dispatcher)


def test_add_router_duplicate():
    router = OnMessage()
    nested = OnMessage()
    router.add_router(nested)
    router.add_router(nested)
    assert len(router._routers) == 1
    assert nested in router._routers


@pytest.mark.asyncio
async def test_resolve_no_handlers(router, update_info):
    router._on_registration()
    assert not await router._resolve(update_info)


@pytest.mark.asyncio
async def test_resolve_catch_all(router, update_info, handler_func):
    router.add_handler(handler_func)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_resolve_none(router, update_info, handler_func):
    filter = Command("foo")
    router.add_handler(handler_func, filters=[filter])
    router._on_registration()
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_resolve_nested_vert(update_info, handler_func):

    def find_router(router):
        r = router
        while r._routers:
            r = r._routers[-1]
        return r

    router = OnMessage()
    for _ in range(5):
        find_router(router).add_router(OnMessage())

    filter = Command("foo")
    update_info.content.text = "/foo"
    find_router(router).add_handler(handler_func, filters=[filter])
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_resolve_nested_horiz(update_info, handler_func):
    router = OnMessage()
    for _ in range(5):
        router.add_router(OnMessage())
    filter = Command("foo")
    update_info.content.text = "/foo"
    router._routers[-1].add_handler(handler_func, filters=[filter])
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_guard_false(handler_func, router, update_info, asyncdef):
    router.add_guard(asyncdef(False))
    router.add_handler(handler_func)
    router._on_registration()
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_guard_true(handler_func, router, update_info, asyncdef):
    router.add_guard(asyncdef(True))
    router.add_handler(handler_func)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_parent_guard_true(handler_func, update_info, asyncdef):
    router = OnMessage()
    router.add_guard(asyncdef(True))
    router.add_handler(handler_func)

    router1 = OnMessage()
    router1.add_handler(asyncdef())

    router.add_router(router1)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_parent_guard_false(handler_func, update_info, asyncdef):
    router = OnMessage()
    router.add_guard(asyncdef(False))
    router.add_handler(asyncdef())

    router1 = OnMessage()
    router1.add_handler(handler_func)

    router.add_router(router1)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_guard_false_both(handler_func, update_info, asyncdef):
    router = OnMessage()
    router.add_guard(asyncdef(False))
    router.add_handler(asyncdef())

    router1 = OnMessage()
    router1.add_guard(asyncdef(False))
    router1.add_handler(handler_func)

    router.add_router(router1)
    router._on_registration()
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_guard_skip_nested(handler_func, update_info, asyncdef):
    router = OnMessage(skip_with_nested=True)
    router.add_guard(asyncdef(False))

    router1 = OnMessage()
    router1.add_guard(asyncdef(True))
    router1.add_handler(handler_func)

    router.add_router(router1)
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


# yapf: disable
def middleware_factory(mark):
    def middleware(handler):
        async def wrapper(update):
            mark()
            await handler(update)
        return wrapper
    return middleware
# yapf: enable


@pytest.mark.asyncio
async def test_resolve_wrap_router_middlewares(
    router, update_info, handler_func
):
    m = mock.Mock()
    router.add_handler(handler_func)
    for _ in range(5):
        router.add_middleware(
            Middleware(
                middleware_factory(m),
                is_handler=True,
            )
        )
    assert len(router._middleware) == 5

    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once_with(update_info.content)
    assert m.call_count == 5


@pytest.mark.asyncio
async def test_resolve_wrap_handler_middlewares(
    router, handler_func, update_info
):
    m = mock.Mock()
    router.add_handler(
        handler_func,
        middleware=[middleware_factory(m) for _ in range(5)],
    )
    router._on_registration()
    assert await router._resolve(update_info)
    assert m.call_count == 5
    handler_func.assert_awaited_once_with(update_info.content)


@pytest.mark.asyncio
async def test_resolve_wrap_middlewares(router, update_info, handler_func):
    m = mock.Mock()
    router.add_handler(
        handler_func,
        middleware=[middleware_factory(m) for _ in range(5)],
    )
    for _ in range(5):
        router.add_middleware(
            Middleware(
                middleware_factory(m),
                is_handler=True,
            )
        )
    router._on_registration()
    assert await router._resolve(update_info)
    assert m.call_count == 10
    handler_func.assert_awaited_once_with(update_info.content)


def test_sort_filters_router_true():
    user = User(123)
    channel = Channel()
    command = Command("start")
    content = Content("text")

    router = OnMessage()
    router.add_handler(
        object(), filters=[
            content,
            command,
            user,
            channel,
        ]
    )
    assert router._handlers[0]._filters == [channel, user, content, command]


def test_sort_filters_router_true_func_false():
    user = User(123)
    channel = Channel()
    command = Command("start")
    content = Content("text")

    router = OnMessage()
    router.add_handler(
        object(),
        filters=[
            content,
            command,
            user,
            channel,
        ],
        sort_filters=False,
    )
    assert router._handlers[0]._filters == [content, command, user, channel]


def test_sort_filters_router_false():
    user = User(123)
    channel = Channel()
    command = Command("start")
    content = Content("text")

    router = OnMessage(sort_filters=False)
    router.add_handler(
        object(), filters=[
            content,
            command,
            user,
            channel,
        ]
    )
    assert router._handlers[0]._filters == [content, command, user, channel]


def test_sort_filters_router_false_func_true():
    user = User(123)
    channel = Channel()
    command = Command("start")
    content = Content("text")

    router = OnMessage(sort_filters=False)
    router.add_handler(
        object(),
        filters=[
            content,
            command,
            user,
            channel,
        ],
        sort_filters=True,
    )
    assert router._handlers[0]._filters == [channel, user, content, command]
